"""
1 AP, 2 STA, hidden node experiment

This experiments demonstrates the effects of the RTS/CTS mechanism of 802.11
on the classic hidden node problem. Note, in order for the results of this
experiment to make sense, nodes should be placed such that STA 1 is hidden
from STA 2.

Flow Legend:

Flow 1) STA 1 ->  AP
Flow 2) STA 2 ->  AP

Experiment Legend:

Exp 0)  RTS/CTS Disabled  -- Timeline: | Flow 1 | Flow 2 | Flow 1 + Flow 2 |
Exp 1)  RTS/CTS Enabled   -- Timeline  | Flow 1 | Flow 2 | Flow 1 + Flow 2 |

"""
import sys
import time
import wlan_exp.config as wlan_exp_config
import wlan_exp.util as wlan_exp_util
import wlan_exp.ltg as wlan_exp_ltg
import wlan_exp.log.util_hdf as hdf_util


#-------------------------------------------------------------------------
#  Global experiment variables
#

# NOTE: change these values to match your experiment setup
NETWORK           = '10.0.0.0' 
NODE_SERIAL_LIST  = ['W3-a-00183', 'W3-a-00381', 'W3-a-00611']

# Set the per-trial duration (in seconds)
TRIAL_TIME = 30

# Set the packet length of each LTG transmission
PACKET_LEN = 1400

# Set the channel for all nodes
CHANNEL = 1

# SSID of the BSS we will create
AP_SSID           = "WARP Hidden Node Exp"

# Set the transmission power in dBm
TX_POW = 10

#-----------------------------------------------------------------------------
# Local Helper Utilities
#-----------------------------------------------------------------------------

def write_log_file(file_name, data_buffer):
    """Writes log data to a HDF5 file."""
    try:        
        # Get the byte log_data out of the WARPNet buffer
        data = data_buffer.get_bytes()
        
        # Write the byte Log_data to the file 
        hdf_util.log_data_to_hdf5(data, file_name, overwrite=True)
    except AttributeError as err:
        print("Error writing log file: {0}".format(err))
        
        
def reset_everything():    
    
    # Tell each STA to not automatically adjust its timebase according to received
    # beacons from the AP. This isn't strictly required, but it will keep the 
    # beginning of the logs cleaner by preventing any race condition between a
    # STA updating itself and receiving the broadcast_cmd_set_time message.
    n_sta_1.sta_configure(beacon_ts_update=True)    
    n_sta_2.sta_configure(beacon_ts_update=True)       
    
    wlan_exp_util.broadcast_cmd_set_time(0.0, network_config)     
     
    # Put each node in a known, good state
    for node in nodes:
        node.set_tx_power(TX_POW)
        node.log_configure(log_full_payloads=False,  log_txrx_ctrl=False, log_txrx_mpdu=True)
        node.reset_all()
    
    # Tell each STA to automatically adjust its timebase according to received
    # beacons from the AP
    n_sta_1.sta_configure(beacon_ts_update=True)    
    n_sta_2.sta_configure(beacon_ts_update=True)    
    
    # Set the network SSID
    n_ap.set_ssid(AP_SSID)
    
    # Force the association state between the AP and STA    
    n_ap.add_association(n_sta_1)
    n_ap.add_association(n_sta_2)
    
    # Check that the nodes are associated.  Otherwise, the LTGs below will fail.
    if not (n_ap.is_associated(n_sta_1) and n_ap.is_associated(n_sta_2)):
        print("\nERROR: Nodes are not associated.")
        print("    Ensure that the AP and the STA are associated.")
        sys.exit(0)
    
  
        
    
def print_and_log(string, file_name):
    f = open(file_name,'a');
    f.write(string)
    f.write("\n")
    f.close()
    print(string)
    


#-------------------------------------------------------------------------
#  Initialization
#
print("\nInitializing experiment\n")

# Create an object that describes the configuration of the host PC
network_config = wlan_exp_config.WlanExpNetworkConfiguration(network=NETWORK) 

# Create an object that describes the WARP v3 nodes that will be used in this experiment
nodes_config   = wlan_exp_config.WlanExpNodesConfiguration(network_config=network_config, 
                                                           serial_numbers=NODE_SERIAL_LIST) 

# Initialize the Nodes
#  This command will fail if either WARP v3 node does not respond
nodes = wlan_exp_util.init_nodes(nodes_config, network_config)

# Extract the AP and STA nodes from the list of initialized nodes
n_ap_l  = wlan_exp_util.filter_nodes(nodes, mac_high='AP')
n_sta_l = wlan_exp_util.filter_nodes(nodes, mac_high='STA')

# Check that we have a valid AP and two STAs
if (((len(n_ap_l) == 1) and (len(n_sta_l) == 2))):
    # Extract the two nodes from the lists for easier referencing below
    n_ap = n_ap_l[0]
    n_sta_1 = n_sta_l[0]
    n_sta_2 = n_sta_l[1]
else:
    print("ERROR: Node configurations did not match requirements of script.\n")
    print(" Ensure two nodes are ready, one using the AP design, one using the STA design\n")
    sys.exit(0)

n_ap.set_channel(CHANNEL)
n_sta_1.set_channel(CHANNEL)
n_sta_2.set_channel(CHANNEL)

#%%
#-------------------------------------------------------------------------
#  Run Experiments
#
print("\nRun Experiment:")


for rate_idx in [2]:

    rate = wlan_exp_util.wlan_rates[rate_idx] 
    n_ap.set_tx_rate_unicast(rate, curr_assoc=True, new_assoc=True)    
    n_sta_1.set_tx_rate_unicast(rate, curr_assoc=True, new_assoc=True)    
    n_sta_2.set_tx_rate_unicast(rate, curr_assoc=True, new_assoc=True)    
    
    for exp_idx in range(2):                        
        print("\nExperiment {0}:".format(exp_idx))
        
        reset_everything()
        
        time.sleep(2)

        if exp_idx == 0:        
            n_ap.set_dcf_rts_thresh(2000)    #Disable RTS/CTS   
            n_sta_1.set_dcf_rts_thresh(2000) #Disable RTS/CTS   
            n_sta_2.set_dcf_rts_thresh(2000) #Disable RTS/CTS   
        elif exp_idx == 1:       
            n_ap.set_dcf_rts_thresh(0)      #Enable RTS/CTS   
            n_sta_1.set_dcf_rts_thresh(0)   #Enable RTS/CTS   
            n_sta_2.set_dcf_rts_thresh(0)   #Enable RTS/CTS            
            

        # Experiment Phase 1              
        flow_1_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
        #flow_2_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)                  
        
        # Wait for the TRIAL_TIME
        time.sleep(TRIAL_TIME)
                
        # Stop all LTG traffic at each node    
        n_sta_1.ltg_stop_all()
        #n_sta_2.ltg_stop_all()
        
        n_sta_1.ltg_remove_all()
        #n_sta_2.ltg_remove_all()
                
        n_sta_1.queue_tx_data_purge_all()
        #n_sta_2.queue_tx_data_purge_all()
        
        # Experiment Phase 2              
        #flow_1_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
        flow_2_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)                  
        
        # Wait for the TRIAL_TIME
        time.sleep(TRIAL_TIME)
                
        # Stop all LTG traffic at each node    
        #n_sta_1.ltg_stop_all()
        n_sta_2.ltg_stop_all()
        
        #n_sta_1.ltg_remove_all()
        n_sta_2.ltg_remove_all()
                
        #n_sta_1.queue_tx_data_purge_all()
        n_sta_2.queue_tx_data_purge_all()
        
        # Experiment Phase 3              
        flow_1_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
        flow_2_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)                  
        
        # Wait for the TRIAL_TIME
        time.sleep(TRIAL_TIME)
                
        # Stop all LTG traffic at each node    
        n_sta_1.ltg_stop_all()
        n_sta_2.ltg_stop_all()
        
        n_sta_1.ltg_remove_all()
        n_sta_2.ltg_remove_all()
                
        n_sta_1.queue_tx_data_purge_all()
        n_sta_2.queue_tx_data_purge_all()                
        
        time.sleep(2)        
        
        write_log_file("E{0}_R{1:>.0f}_sta_1.hdf5".format(exp_idx, rate['rate']), n_sta_1.log_get_all_new(log_tail_pad=0))
        write_log_file("E{0}_R{1:>.0f}_sta_2.hdf5".format(exp_idx, rate['rate']), n_sta_2.log_get_all_new(log_tail_pad=0))
        write_log_file("E{0}_R{1:>.0f}_ap.hdf5".format(exp_idx, rate['rate']), n_ap.log_get_all_new(log_tail_pad=0))            
