import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import log_analysis_util

PKT_TYPE_LTG = 3
PKT_TYPE_MGMT = 11

T_MIN = 0E6
T_MAX = 95E6

#Define nodes - each entry must be 2-tuple:
# (node_name, HDF5_raw_log_file_name)
nodes = (
    ('AP',    'E0_R12_ap.hdf5'),
    ('STA 1', 'E0_R12_sta_1.hdf5'),
    ('STA 2', 'E0_R12_sta_2.hdf5'),
)

flow_conds_common = (
    ('timestamp', '>', T_MIN),
    ('timestamp', '<', T_MAX),
)

flow_cond_ltg = ('pkt_type', '==', PKT_TYPE_LTG)

#Define flows - each entry must be 4-tuple:
# (src, dest, conds_tx, conds_rx)
# src/dest: either string name matching name in nodes list or MAC address as uint64
# conds: list of 3-tuples (field_name, operator, value); use [] for no conditions
flows = (
    ('STA 1', 'AP',    [flow_cond_ltg], [flow_cond_ltg]),
    ('STA 2', 'AP',    [flow_cond_ltg], [flow_cond_ltg])
)

#Plotting colors per flow - each entry must be 3-tuple:
# (color_Tx, color_Rx_good, color_Rx_bad)
# Each color is 4-bytes as u32: (alpha, red, green, blue)
flow_colors = (
    (0xFFFF0000, 0xA0FF0000, 0xA0AA0000), #Red
    (0xFF00CC00, 0xA000CC00, 0xA0007700), #Green
    (0xFFFF9500, 0xA0FF9500, 0xA0AA4000), #Orange
    (0xFF0C5DA5, 0xA00C5DA5, 0xA0000850), #Blue
    (0xFF000000, 0xFF000000, 0xFF222222), #Black
)

#%%

#Extract Tx/Rx entries per flow per node
flows_tx, flows_rx, all_entries_np = log_analysis_util.proc_txrx_logs_to_flows(nodes, flows, flow_conds_common)

#%%


##########################################
# Plot #1: Rx throughput vs. time per flow
#  Process:
#   1) Extract times and Rx lengths from destination node of each flow
#   2) Resample Rx lengths with regular interval
#   3) Merge time indexes of resampled Rx length vectors
#   4) Compute rolling mean of each resampled Rx length vector
#   5) Scale length vs. time vector to be throughput in Mbps
#   6) Plot individual flow throughputs and their sum

#Resample interval (miliseconds)
rs_interval = 1

#Rolling window (# of samples per window)
rolling_window = 1000

#Itereate over flows to extract and resample Rx length vectors from destination 
# node for each flow 
flows_rx_lens_rs = list()
for f_idx,flow in enumerate(flows):

    f_src = flow[0]
    f_dst = flow[1]

    #Extract all Rx entries from target node
    f_rx = flows_rx[f_idx][ [n[0] for n in nodes].index(f_dst) ]

    #Select good-checksum, non-duplicated receptions
    f_rx_idx = ((f_rx['flags'] & 0x1) == 0) & (f_rx['fcs_result'] == 0)
    f_rx = f_rx[f_rx_idx]
    
    f_rx_t    = f_rx['timestamp']
    f_rx_lens = f_rx['length']

    #Insert leading and trailing zeros, so the rolling window below
    # has valid data on both sides of the actual data
    f_rx_t    = np.insert(f_rx_t, 0, 0)
    f_rx_t    = np.insert(f_rx_t, len(f_rx_t)-1, np.max(f_rx_t)+(rolling_window*rs_interval*1000))

    f_rx_lens = np.insert(f_rx_lens, 0, 0)
    f_rx_lens    = np.insert(f_rx_lens, len(f_rx_lens)-1, 0)
    
    #Convert Rx timestamps to Pandas datetime index
    f_rx_t_pd = pd.to_datetime(f_rx_t, unit='us')

    #Convert Rx pkt length array to Pandas series indexed by Rx timestamps
    f_rx_lens_pd = pd.Series(f_rx_lens, index=f_rx_t_pd)  
    
    #Resample Rx length series to regular sampling interval
    f_rx_lens_pd_rs = f_rx_lens_pd.resample(('%dL' % rs_interval), how='sum').fillna(value=0)

    #Remember this flow's resampled Rx lengths for later plotting
    flows_rx_lens_rs.insert(f_idx, f_rx_lens_pd_rs)

    #Construct the running union of all flow's resampled Rx timestamps
    if(f_idx == 0):
        flows_rx_t_idx = f_rx_lens_pd_rs.index
    else:
        flows_rx_t_idx = flows_rx_t_idx.union(f_rx_lens_pd_rs.index)

#Itereave over resampled Rx length vectors to merge indexes and calculate mean throughput
flow_xputs = list()
for f_idx in range(len(flows_rx_lens_rs)):
    #Re-index each flow's Rx lengths series by the common-to-all-flows time index
    flows_rx_lens_rs[f_idx] = flows_rx_lens_rs[f_idx].reindex(flows_rx_t_idx)

    #Compute the rolling mean of the Rx lengths
    f_xput = pd.rolling_mean(flows_rx_lens_rs[f_idx], window=rolling_window, min_periods=1).fillna(value=0)

    #Scale the Rx lengths to throughputs
    # 1.0e-6: Convert microseconds to seconds
    # 8.0: Convert bytes to bits
    # (1.0/(rs_interval * 1e-3): Convert resampling interval in msec to seconds
    f_xput = f_xput  * (1.0e-6 * 8.0 * (1.0/(rs_interval * 1e-3)))

    #Remember each flow's xput for plotting
    flow_xputs.insert(f_idx, f_xput)


#Create time vector for plot X-axis, in units of seconds
t_sec = flows_rx_t_idx.astype('int64') / 1.0E9
plt_t = t_sec

#Open a clean figure for plotting
plt.close(1)
fig = plt.figure(1, figsize=(10,8))

#Plot each throughput vector and their sum
sum_xput = flow_xputs[0]
for f_idx,f_xput in enumerate(flow_xputs):
    if(f_idx > 0):
        sum_xput += f_xput

    plt.plot(plt_t, f_xput, 
        color='#%06x' % (0xFFFFFF & flow_colors[f_idx][0]), #Color as HTML
        label='Flow {0}'.format(f_idx+1))

plt.plot(plt_t, sum_xput, color='#000000', label='Sum of Flows')

#Place a legend above the axes
leg = plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=8,
       ncol=3, borderaxespad=0.)
       
#Make the legend line widths thicker
for legobj in leg.legendHandles:
    legobj.set_linewidth(5.0)


#Label and format the axes
plt.xlabel('Time (sec)')
plt.ylabel('Throughput (Mb/sec)')
plt.xlim([(T_MIN/1E6) + 5, (T_MAX/1E6) - 5])
plt.grid('on')
plt.axis([0, 95, 0, 12])

plt.savefig('xput_vs_time.png')
